try:
    import importlib_metadata as meta
except ImportError:
    import importlib.metadata as meta  # type: ignore

__title__ = "mckit"
__distribution__ = meta.distribution(__title__)
__meta_data__ = __distribution__.metadata
__author__ = __meta_data__["Author"]
__author_email__ = __meta_data__["Author-email"]
__license__ = __meta_data__["License"]
__summary__ = __meta_data__["Summary"]
__copyright__ = (
    "Copyright 2018-2020 Roman Rodionov"  # TODO dvp: move to meta (project.toml)
)
__version__ = __distribution__.version


#
# The version from metadata may have several formats:
#  - in release version (there are no prepatch suffixes) it will be plain 1.5.0 for example
#  - in prepatch version running separately 1.5.0-alpha1
#  - in prepatch versions test runs it will be 1.5.0a1
#
# VERSION_PATTERN = re.compile(
#     r"""
#     ^
#     v?
#     (?:
#         (?:(?P<epoch>[0-9]+)!)?                           # epoch
#         (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
#         (?P<pre>                                          # pre-release
#             [-_.]?
#             (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
#             [-_.]?
#             (?P<pre_n>[0-9]+)?
#         )?
#         (?P<post>                                         # post release
#             (?:-(?P<post_n1>[0-9]+))
#             |
#             (?:
#                 [-_.]?
#                 (?P<post_l>post|rev|r)
#                 [-_.]?
#                 (?P<post_n2>[0-9]+)?
#             )
#         )?
#         (?P<dev>                                          # dev release
#             [-_.]?
#             (?P<dev_l>dev)
#             [-_.]?
#             (?P<dev_n>[0-9]+)?
#         )?
#     )
#     (?:\+(?P<local>[a-z0-9]+(?:[-_.][a-z0-9]+)*))?       # local version
#     $
# """,
#     re.IGNORECASE | re.VERBOSE,
# )



# -*- coding: utf-8 -*-

from functools import reduce

from .printer import print_card, separate


class Distribution:
    """Represents a distribution of source variable.

    If the len of probs equals to the len of values, then the distribution is
    either of discrete variable (if is_discrete flag is set) or probs sets
    probability density values. If len(probs) + 1 == len(values), then
    values are bin boundaries for continuous distribution.

    Parameters
    ----------
    name : int
        Distribution's name. It is needed for MCNP.
    values : list
        A list of values, that variable can be equal to. Values can be float,
        int and other Distribution instances.
    probs : list or Distribution
        A list of probabilities or a distribution this one depends on.
        Probabilities need not be normalized.
    variable : str
        Source variable's name. Default: None.
    is_discrete : bool
        Indicate that the variable is discrete. Default: False.

    Methods
    -------
    mcnp_repr()
        Returns a string representation of corresponding MCNP card.
    get_inner()
        Gets all distributions this distribution depends on.
    depends_on()
        Gets distribution this one depends on.

    Properties
    ----------
    name : int
        Distribution's name.
    size : int
        Size of the distribution. It is the number of values or bins.
    variable : str
        Name of source variable.
    """

    def __init__(self, name, values, probs, variable=None, is_discrete=False):
        self._name = name
        self._var = variable
        Distribution.check_distr(values, len(probs), is_discrete)
        self._is_discrete = is_discrete
        self._values = list(values)
        if isinstance(probs, Distribution):
            self._probs = probs
        else:
            self._probs = list(probs)
        if (
            isinstance(self._probs, list)
            and len(self._probs) == len(values)
            and not is_discrete
            and not isinstance(self._values[0], Distribution)
        ):
            self._is_pdf = True
        else:
            self._is_pdf = False

    @property
    def name(self):
        """Gets distribution's name."""
        return self._name

    @property
    def size(self):
        """Gets distribution's size."""
        return len(self._probs)

    @property
    def variable(self):
        """Gets distribution's source variable name."""
        return self._var

    def mcnp_repr(self):
        """Returns a string representation of corresponding MCNP card."""
        tokens = []
        discrete = self._is_discrete
        is_pdf = self._is_pdf
        if isinstance(self._values[0], Distribution):
            tokens.append("S")
            for v in self._values:
                tokens.append(str(v.name))
        else:
            if isinstance(self._values[0], list):
                values = reduce(list.__add__, self._values)
            else:
                values = self._values
            if discrete:
                tokens.append("L")
            elif is_pdf:
                tokens.append("A")
            else:
                tokens.append("H")
            for v in values:
                tokens.append(str(v))

        if isinstance(self._probs, Distribution):
            tokens.insert(0, "DS{0}".format(self._name))
            prob_tokens = None
        else:
            tokens.insert(0, "SI{0}".format(self._name))
            prob_tokens = ["SP{0}".format(self._name)]
            if not is_pdf:
                prob_tokens.append("D")
            if len(self._values) == len(self._probs) + 1:
                prob_tokens.append("0")
            for p in self._probs:
                prob_tokens.append(str(p))
        card = print_card(separate(tokens))
        if prob_tokens:
            card += "\n" + print_card(separate(prob_tokens))
        return card

    def get_inner(self):
        """Gets nested distributions this one depends on.

        If values of this distribution are distributions itself, then they
        are returned.

        Returns
        -------
        dists : set
            A set of nested distributions.
        """
        dists = set()
        if isinstance(self._values[0], Distribution):
            for v in self._values:
                dists.add(v)
        return dists

    def __len__(self):
        return len(self._probs)

    def depends_on(self):
        """Gets distribution this one depends on.

        Returns
        -------
        dist : Distribution
            Distribution this one depends on. None if the distribution is
            independent.
        """
        if isinstance(self._probs, Distribution):
            return self._probs
        else:
            return None

    @staticmethod
    def check_distr(values, size, is_discrete):
        """Checks if the distribution is correct.

        Parameters
        ----------
        values : array_like
            List of variable values.
        size : int
            The length of intensity matrix along variable dimension.
        is_discrete : bool
            If True, the distribution is assumed to be discrete.
        """
        discr_or_pdf = len(values) == size
        histogram = len(values) == size + 1 and not is_discrete
        if not discr_or_pdf and not histogram:
            raise ValueError("Inconsistent size of values.")


def create_bin_distributions(bins, start_name=1):
    """Creates bin distributions for specified bins.

    Parameters
    ----------
    bins : array_like
        A list of bin boundaries.
    start_name : int
        Starting name for distributions. For every new distribution the name
        is incremented by 1.

    Returns
    -------
    free_name : int
        Distribution name that can be used for new distributions.
    distributions : list
        A list of distributions created. Index in the list corresponds to the
        index of bin.
    """
    distributions = []
    for low, high in zip(bins[:-1], bins[1:]):
        distributions.append(Distribution(start_name, [low, high], [1]))
        start_name += 1
    return start_name, distributions


def expand_matrix_distribution(intensities, *var_values, start_name=1):
    """Converts matrix distribution to the len(var_values) linear.

    Parameters
    ----------
    intensities : array_like
        A matrix of source intensities.
    var_values : tuple
        A tuple of variable values along each axis. Length of var_values must
        be equal to the number of intensities dimensions.
    start_name : int
        Starting name for distributions. Default: 1.

    Returns
    -------
    free_name : int
        The name that can be used for new distributions.
    exp_var_values : tuple
        A tuple of lists of values for each source variable.
    exp_intensities : list[float]
        A list of expanded intensities for every hyperbin.
    """
    if len(intensities.shape) != len(var_values):
        raise ValueError("Inconsistent number of variables")
    uniq_values = []
    exp_var_values = []
    for dim, values in zip(intensities.shape, var_values):
        if Distribution.check_distr(values, dim):
            uniq_values.append(values)
        else:
            start_name, dists = create_bin_distributions(values, start_name)
            uniq_values.append(dists)
        exp_var_values.append([])

    exp_intensities = []
    for index, intensity in intensities:
        for j, i in enumerate(index):
            exp_var_values[j].append(uniq_values[j][i])
        exp_intensities.append(intensity)
    return start_name, tuple(exp_var_values), exp_intensities


class Source:
    """Represents particle source.

    Parameters
    ----------
    variables : dict
        A dictionary of source variables. var_name -> var_value. The value
        can be either float (int) value or Distribution instance.

    Methods
    -------
    mcnp_repr()
        Gets a string representation of corresponding MCNP card.
    """

    def __init__(self, **variables):
        self._variables = variables

    def mcnp_repr(self):
        """Gets a string representation of corresponding MCNP card."""
        tokens = ["SDEF"]
        cards = []
        extra_cards = []
        for k, v in self._variables.items():
            tokens.append("{0}={1}".format(k, Source._var_repr(k, v)))
            if isinstance(v, Distribution):
                cards.append(v.mcnp_repr())
                for ec in sorted(v.get_inner(), key=lambda x: x.name):
                    extra_cards.append(ec.mcnp_repr())
        cards.insert(0, print_card(separate(tokens)))
        cards.extend(extra_cards)
        return "\n".join(cards)

    @staticmethod
    def _var_repr(key, value):
        if isinstance(value, Distribution):
            dep = value.depends_on()
            result = "D{0}".format(value.name)
            if dep:
                result = "F{0} ".format(dep.variable) + result
        else:
            result = str(value)
        return result



# -*- coding: utf-8 -*-
"""
 Read and parse MCNP file text.
"""
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    List,
    NewType,
    Optional,
    TextIO,
    Tuple,
    Union,
)

from itertools import repeat
from pathlib import Path

from attr import attrib, attrs
from mckit.card import Card
from mckit.constants import MCNP_ENCODING
from mckit.parser.cell_parser import Body
from mckit.parser.cell_parser import parse as parse_cell
from mckit.parser.common import (
    CellNotFoundError,
    CellStrictIndex,
    CompositionStrictIndex,
    ParseError,
    SurfaceStrictIndex,
    TransformationStrictIndex,
)
from mckit.parser.material_parser import Composition
from mckit.parser.material_parser import parse as parse_composition
from mckit.parser.surface_parser import Surface
from mckit.parser.surface_parser import parse as parse_surface
from mckit.parser.transformation_parser import Transformation
from mckit.parser.transformation_parser import parse as parse_transformation
from mckit.universe import Universe, produce_universes
from mckit.utils.Index import Index

from .mcnp_section_parser import Card as TextCard
from .mcnp_section_parser import (
    InputSections,
    Kind,
    distribute_cards,
    parse_sections_text,
)

T = NewType("T", Any)
T1 = NewType("T1", Any)
T2 = NewType("T2", Any)
Pair = NewType("Pair", Tuple[T1, T2])
YieldGenerator = NewType("YieldGenerator", Generator[T, None, None])


@attrs
class ParseResult:
    universe: Universe = attrib()
    cells: List[Body] = attrib()
    cells_index: CellStrictIndex = attrib()
    surfaces: List[Surface] = attrib()
    surfaces_index: SurfaceStrictIndex = attrib()
    compositions: Optional[List[Composition]] = attrib()
    compositions_index: Optional[CompositionStrictIndex] = attrib()
    transformations: Optional[List[Transformation]] = attrib()
    transformations_index: Optional[TransformationStrictIndex] = attrib()
    sections: InputSections = attrib()

    @property
    def title(self):
        return self.sections.title


def from_file(path: Union[str, Path]) -> ParseResult:
    if isinstance(path, str):
        path = Path(path)
    with path.open("r", encoding=MCNP_ENCODING) as fid:
        return from_stream(fid)


def from_stream(stream: TextIO) -> ParseResult:
    text = stream.read()
    return from_text(text)


def from_text(text: str) -> ParseResult:
    sections: InputSections = parse_sections_text(text)
    if sections.data_cards:
        # fmt: off
        text_compositions, text_transformations, _, _, _ = distribute_cards(
            sections.data_cards
        )  # type: List[TextCard], List[TextCard], List[TextCard], List[TextCard], List[TextCard],
        # fmt: on
        transformations = parse_transformations(text_transformations)
        transformations_index = TransformationStrictIndex.from_iterable(transformations)
        compositions = parse_compositions(text_compositions)
        compositions_index = CompositionStrictIndex.from_iterable(compositions)
    else:
        transformations = None
        transformations_index = None
        compositions = None
        compositions_index = None

    surfaces = parse_surfaces(sections.surface_cards, transformations_index)
    surfaces_index = SurfaceStrictIndex.from_iterable(surfaces)
    cells, cells_index = parse_cells(
        sections.cell_cards, surfaces_index, compositions_index, transformations_index
    )
    universe = produce_universes(
        cells
    )  # TODO dvp: apply title from sections to the topmost universe
    return ParseResult(
        universe=universe,
        cells=cells,
        cells_index=cells_index,
        surfaces=surfaces,
        surfaces_index=surfaces_index,
        compositions=compositions,
        compositions_index=compositions_index,
        transformations=transformations,
        transformations_index=transformations_index,
        sections=sections,
    )


TSectionGenerator = NewType(
    "TSectionGenerator",
    Generator[Union[Transformation, Composition, Surface], None, None],
)


def join_comments(text_cards: Iterable[TextCard]):
    def _iter():
        comment: Optional[str] = None
        for card in text_cards:
            if card.is_comment:
                assert comment is None, f"Comment is already set {comment[:70]}"
                comment = card.text
            else:
                assert (
                    not card.is_comment
                ), f"Pair of comment is found, second one is: {comment[:70]}"
                yield card, comment
                comment = None

    return list(_iter())


def parse_section(
    text_cards: Iterable[TextCard], expected_kind: Kind, parser: Callable[[str], Card]
) -> TSectionGenerator:
    text_cards_with_comments = join_comments(text_cards)

    def iterator() -> TSectionGenerator:
        for text_card, comment in text_cards_with_comments:
            assert text_card.kind is expected_kind
            try:
                card = parser(text_card.text)
            except (ValueError, ParseError) as ex:
                raise ValueError(f"Failed to parse card '{text_card}'") from ex
            if comment:
                card.options["comment_above"] = comment
            # card.options['original'] = text_card.text
            yield card

    return iterator()


def parse_transformations(text_cards: Iterable[TextCard]) -> List[Transformation]:
    return list(parse_section(text_cards, Kind.TRANSFORMATION, parse_transformation))


def parse_compositions(text_cards: Iterable[TextCard]) -> List[Composition]:
    return list(parse_section(text_cards, Kind.MATERIAL, parse_composition))


def parse_surfaces(
    text_cards: Iterable[TextCard], transformations: Index
) -> List[Surface]:
    def parser(text: str):
        return parse_surface(text, transformations=transformations)

    return list(parse_section(text_cards, Kind.SURFACE, parser))


def extract_number(text_card: TextCard):
    return int(text_card.text.split(maxsplit=1)[0])


class MissedCellsError(RuntimeError):
    def __init__(self, missed_cells: List[int]):
        self.missed_cells = missed_cells
        msg = f"Not found cells: {missed_cells}"
        super().__init__(msg)

    @classmethod
    def from_text_cards(cls, missed_cells: Iterable[TextCard]):
        missed_cells_numbers: List[int] = list(map(extract_number, missed_cells))
        return cls(missed_cells_numbers)


def parse_cells(
    text_cards: List[TextCard],
    surfaces: Index,
    compositions: Index,
    transformations: Index,
) -> Tuple[List[Body], CellStrictIndex]:
    text_cards_with_comments = join_comments(text_cards)
    size = len(text_cards_with_comments)
    cells_index = CellStrictIndex()
    cells_to_process = list(range(size))
    cells = list(repeat(None, size))

    def parser(text: str):
        return parse_cell(
            text,
            cells=cells_index,
            surfaces=surfaces,
            compositions=compositions,
            transformations=transformations,
        )

    while cells_to_process:
        cells_to_process_length = len(cells_to_process)
        new_cells_to_process = []
        for i in cells_to_process:
            text_card, comment = text_cards_with_comments[i]
            assert text_card.kind is Kind.CELL
            try:
                card: Body = parser(text_card.text)
                assert card is not None, (
                    "Failed to process cell %s" % text_card.text[:70]
                )
            except CellNotFoundError:
                new_cells_to_process.append(i)
                continue
            if comment:
                card.options["comment_above"] = comment
            card.options["original"] = text_card.text
            cells[i] = card
            cells_index[card.name()] = card
        if cells_to_process_length == len(new_cells_to_process):
            missed_cells_cards = list(text_cards[i] for i in new_cells_to_process)
            raise MissedCellsError.from_text_cards(missed_cells_cards)
        cells_to_process = new_cells_to_process

    if any(map(lambda c: c is None, cells)):
        cells = list(filter(lambda c: c is not None, cells))

    return cells, cells_index



"""Features, common for all cards"""
from typing import List, Optional, Text

from abc import ABC, abstractmethod
from functools import reduce
from operator import xor

from mckit.utils.named import Name

from .printer import print_card


class Card(ABC):
    """Features, common for all cards."""

    def __init__(self, **options):
        self.options = options

    def name(
        self,
    ) -> Optional[
        Name
    ]:  # TODO dvp: we'd better have special property name, don't use options for that
        """Returns card's name."""
        return self.options.get("name", None)

    def rename(self, new_name) -> "Card":
        """Renames the card."""
        self.options["name"] = new_name
        return self

    @property
    def is_anonymous(self):
        return self.name() is None

    @abstractmethod
    def mcnp_words(self, pretty=False) -> List[Text]:
        """Gets a list of card words."""

    @property
    def has_original(self) -> bool:
        return "original" in self.options.keys()

    @property
    def has_comment_above(self) -> bool:
        return "comment_above" in self.options.keys()

    def mcnp_repr(self, pretty: bool = False) -> Text:
        """Gets str representation of the card."""
        # TODO dvp: try to use original texts, if available
        # if self.has_original:
        #     if self.has_comment_above:
        #         return self.comment_above + '\n' + self.original
        #     else:
        #         return self.original
        # else:
        text: Text = print_card(self.mcnp_words(pretty))
        return text

    @property
    def comment_above(self) -> Optional[Text]:
        comment: Optional[Text] = self.options.get("comment_above", None)
        return comment

    @property
    def original(self) -> Optional[Text]:
        original: Text = self.options.get("original", None)
        return original

    def drop_original(self) -> None:
        del self.options["original"]

    def add_comment(self, *comment: str) -> None:
        comm = self.options.get("comment", [])
        comm.extend(comment)
        self.options["comment"] = comm

    def __str__(
        self,
    ):  # TODO dvp: option `name` is printed twice, should be explicit property of this class instance
        return '{}: "{}"'.format(self.name(), self.options)

    def __hash__(
        self,
    ):  # TODO dvp: dict hashing implementation: check for effect of instability of the options
        return reduce(xor, (hash(k) ^ hash(v) for k, v in self.options.items()), 0)

    def __eq__(self, other):  # TODO dvp: what about nested dictionaries?
        for k in other.options.keys():
            if k not in self.options:
                return False
        for k, v in self.options.items():
            if k not in other.options:
                return False
            my = self.options[k]
            their = other.options[k]
            if my != their:
                return False
        return True
